@@ -355,7 +355,7 @@ enum ie_rsn_akm_suite {
 	 (akm == IE_RSN_AKM_SUITE_FT_OVER_FILS_SHA384))
 
 #define IE_LEN(ie) \
-	(ie) ? (ie)[1] + 2 : 0
+	((ie) ? (ie)[1] + 2 : 0)
 
 struct ie_tlv_iter {
 	unsigned int max;
